/* 
 * Copyright (C) 2001-2007 Jacek Sieka, arnetheduck on gmail point com
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA.
 */

#include "stdafx.h"
#include "../client/DCPlusPlus.h"
#include "Resource.h"

#include "ADLSProperties.h"
#include "../client/ADLSearch.h"
#include "WinUtil.h"

WinUtil::TextItem ADLSProperties::texts[] = {
	{ IDD_ADLS_PROPERTIES, ResourceManager::ADLSEARCH_PROPERTIES },
	{ IDC_ADLSEARCH_SEARCH_STRING, ResourceManager::ADLSEARCH_SEARCH_STRING },
	{ IDC_ADLSEARCH_SOURCE_TYPE, ResourceManager::ADLSEARCH_SOURCE_TYPE },
	{ IDC_ADLSEARCH_DESTINATION_DIRECTORY, ResourceManager::ADLSEARCH_DESTINATION_DIRECTORY },
	{ IDC_ADLSEARCH_MIN_SIZE, ResourceManager::ADLSEARCH_MIN_SIZE },
	{ IDC_ADLSEARCH_MAX_SIZE, ResourceManager::ADLSEARCH_MAX_SIZE }, 
	{ IDC_IS_ACTIVE, ResourceManager::ADLSEARCH_ACTIVE_SEARCH },
	{ IDC_ADLSEARCH_SIZE_TYPE, ResourceManager::ADLSEARCH_SIZE_TYPE },
	{ IDOK, ResourceManager::OK },
	{ IDCANCEL, ResourceManager::CANCEL },
	{ 0, ResourceManager::SETTINGS_AUTO_AWAY }
};

// Initialize dialog
LRESULT ADLSProperties::OnInitDialog(UINT, WPARAM, LPARAM, BOOL&) {
	WinUtil::translate(*this, texts);
	// Initialize dialog items
	ctrlSearch.Attach(GetDlgItem(IDC_SEARCH_STRING));
	ctrlDestDir.Attach(GetDlgItem(IDC_DEST_DIR));
	ctrlMinSize.Attach(GetDlgItem(IDC_MIN_FILE_SIZE));
	ctrlMaxSize.Attach(GetDlgItem(IDC_MAX_FILE_SIZE));
	ctrlActive.Attach(GetDlgItem(IDC_IS_ACTIVE));
	ctrlAutoQueue.Attach(GetDlgItem(IDC_AUTOQUEUE));

	ctrlSearchType.Attach(GetDlgItem(IDC_SOURCE_TYPE));
	ctrlSearchType.AddString(CTSTRING(FILENAME));
	ctrlSearchType.AddString(CTSTRING(DIRECTORY));
	ctrlSearchType.AddString(CTSTRING(ADL_FULL_PATH));

	ctrlSizeType.Attach(GetDlgItem(IDC_SIZE_TYPE));
	ctrlSizeType.AddString(CTSTRING(B));
	ctrlSizeType.AddString(CTSTRING(KB));
	ctrlSizeType.AddString(CTSTRING(MB));
	ctrlSizeType.AddString(CTSTRING(GB));

	// Load search data
	ctrlSearch.SetWindowText(Text::toT(search->searchString).c_str());
	ctrlDestDir.SetWindowText(Text::toT(search->destDir).c_str());
	ctrlMinSize.SetWindowText(Text::toT(search->minFileSize > 0 ? Util::toString(search->minFileSize) : "").c_str());
	ctrlMaxSize.SetWindowText(Text::toT(search->maxFileSize > 0 ? Util::toString(search->maxFileSize) : "").c_str());
	ctrlActive.SetCheck(search->isActive ? 1 : 0);
	ctrlAutoQueue.SetCheck(search->isAutoQueue ? 1 : 0);
	ctrlSearchType.SetCurSel(search->sourceType);
	ctrlSizeType.SetCurSel(search->typeFileSize);

	// Center dialog
	CenterWindow(GetParent());

	return FALSE;
}

// Exit dialog
LRESULT ADLSProperties::OnCloseCmd(WORD /*wNotifyCode*/, WORD wID, HWND /*hWndCtl*/, BOOL& /*bHandled*/) {
	if(wID == IDOK) {
		// Update search
		TCHAR buf[256];

		ctrlSearch.GetWindowText(buf, 256);
		search->searchString = Text::fromT(buf);
		ctrlDestDir.GetWindowText(buf, 256);
		search->destDir = Text::fromT(buf);

		ctrlMinSize.GetWindowText(buf, 256);
		search->minFileSize = (_tcslen(buf) == 0 ? -1 : Util::toInt64(Text::fromT(buf)));
		ctrlMaxSize.GetWindowText(buf, 256);
		search->maxFileSize = (_tcslen(buf) == 0 ? -1 : Util::toInt64(Text::fromT(buf)));

		search->isActive = (ctrlActive.GetCheck() == 1);
		search->isAutoQueue = (ctrlAutoQueue.GetCheck() == 1);

		search->sourceType = (ADLSearch::SourceType)ctrlSearchType.GetCurSel();
		search->typeFileSize = (ADLSearch::SizeType)ctrlSizeType.GetCurSel();
	}

	EndDialog(wID);
	return 0;
}

/* Some parts Copyright (C) 2002-2008 Petr Kozelka, PPK at czdc dot org */