/* 
 * Copyright (C) 2004 createagaga
 * Copyright (C) 2004-2008 Petr Kozelka, PPK at czdc dot org
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA.
 */

#include "stdafx.h"
#include "../client/DCPlusPlus.h"
#include "AGEmoticonSetup.h"
#include "../client/SimpleXML.h"
#include "../client/pointer.h"

CAGEmoticon::CAGEmoticon() {
	m_EmoticonText = Util::emptyStringT;
	m_EmoticonBmp = NULL;
	bWithAlpha = false;
	lWidth = 0;
	lHeight = 0;
}

CAGEmoticon::~CAGEmoticon() {
	if(m_EmoticonBmp != NULL) {
		DeleteObject(m_EmoticonBmp);
		m_EmoticonBmp = NULL;
	}
}

bool CAGEmoticon::Create(tstring& strEmoticonText, tstring& strEmoticonBmpPath) {
	m_EmoticonBmp = (HBITMAP) ::LoadImage(0, strEmoticonBmpPath.c_str(), IMAGE_BITMAP, 0, 0, LR_LOADFROMFILE | LR_CREATEDIBSECTION);
	if(m_EmoticonBmp == NULL) {
		TCHAR buf[1024];
		_stprintf_s(buf, 1024, _T("Unable to load '%s'."), strEmoticonBmpPath.c_str());
		MessageBox(NULL, buf, _T("Unable to load file"), MB_ICONSTOP | MB_OK);
		return false;
	}

	m_EmoticonText = strEmoticonText;

	BITMAP bm;
	GetObject(m_EmoticonBmp, sizeof(bm), &bm);

	bWithAlpha = (bm.bmBitsPixel == 32);
	lWidth = bm.bmWidth;
	lHeight = bm.bmHeight;
	
	if(bWithAlpha == false) {
        return true;
    }

	AutoArray<BYTE> bpBits(lWidth * lHeight * 4);
	GetBitmapBits(m_EmoticonBmp, lWidth * lHeight * 4, bpBits);

    for(int y = 0; y < lHeight; y++) {
        BYTE * bpPixel = (BYTE *)bpBits + lWidth * 4 * y;

        for(int x = 0; x < lWidth; x++) {
            bpPixel[0] = bpPixel[0] * bpPixel[3] / 255; 
            bpPixel[1] = bpPixel[1] * bpPixel[3] / 255; 
            bpPixel[2] = bpPixel[2] * bpPixel[3] / 255; 

            bpPixel += 4;
        }
    }

    SetBitmapBits(m_EmoticonBmp, lWidth * lHeight * 4, bpBits);

	return true;
}

const tstring& CAGEmoticon::GetEmoticonText() {
	return m_EmoticonText;
}

HBITMAP CAGEmoticon::GetEmoticonBmp(const COLORREF &clrBkColor) {
	HDC DirectDC = CreateCompatibleDC(NULL);
	HDC memDC = CreateCompatibleDC(DirectDC);
	
	HBITMAP DirectBitmap = CreateBitmap(lWidth, lHeight, 1, 32, NULL);

	SelectObject(memDC, m_EmoticonBmp);
	SelectObject(DirectDC, DirectBitmap);

	HBRUSH hbrushBG = CreateSolidBrush(clrBkColor);
	RECT rect = { 0, 0, lWidth, lHeight };
	FillRect(DirectDC, &rect, hbrushBG);
	DeleteObject(hbrushBG);

	if(bWithAlpha == true) {
        BLENDFUNCTION pixelblend = { AC_SRC_OVER, 0, 255, AC_SRC_ALPHA };
		AlphaBlend(DirectDC, 0, 0, lWidth, lHeight, memDC, 0, 0, lWidth, lHeight, pixelblend);
	} else {
		TransparentBlt(DirectDC, 0, 0, lWidth, lHeight, memDC, 0, 0, lWidth, lHeight, GetPixel(memDC, 0, 0));	
	}

	DeleteDC(memDC);
	DeleteDC(DirectDC);

    return DirectBitmap;
}

// CAGEmoticonSetup
CAGEmoticonSetup::CAGEmoticonSetup() {
	setUseEmoticons(false);
}

CAGEmoticonSetup::~CAGEmoticonSetup() {
	for_each(EmoticonsList.begin(), EmoticonsList.end(), DeleteFunction());
}

bool CAGEmoticonSetup::Create(const string& emoticonfile) {
	setUseEmoticons(false);

	if(File::existsFile(emoticonfile) == false)
		return true;

	try {
		SimpleXML xml;
		xml.fromXML(File(emoticonfile, File::READ, File::OPEN).read());
		
		if(xml.findChild("Emotions")) {
			xml.stepIn();

			while(xml.findChild("Emotion")) {
				tstring strEmoticonText = Text::toT(xml.getChildAttrib("ReplacedText"));
				tstring strEmoticonBmpPath = Text::toT(xml.getChildAttrib("BitmapPath"));

				if(strEmoticonBmpPath.size() > 0) {
					strEmoticonBmpPath = Text::toT(Util::getFilePath(emoticonfile)) + strEmoticonBmpPath;
				}

				CAGEmoticon* pEmoticon = new CAGEmoticon();
				if(pEmoticon->Create(strEmoticonText, strEmoticonBmpPath) == false) {
					delete pEmoticon;
					continue;
				}

				EmoticonsList.push_back(pEmoticon);
			}
			xml.stepOut();
		}
	} catch(const Exception& e) {
		dcdebug("CAGEmoticonSetup::Create: %s\n", e.getError().c_str());
		return false;
	}

	if(EmoticonsList.size() == 0) {
		return false;
	}
		
	setUseEmoticons(true);

	return true;
}