/* 
 * Copyright (C) 2004-2008 Petr Kozelka, PPK at czdc dot org
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA.
 */

#include "stdafx.h"
#include "../client/DCPlusPlus.h"
#include "../client/version.h"
#include "Resource.h"
#include "AboutDlg.h"
#include "WinUtil.h"

LRESULT AboutDlg::OnInitDialog(UINT /*uMsg*/, WPARAM /*wParam*/, LPARAM /*lParam*/, BOOL& /*bHandled*/) {
	SetDlgItemText(IDC_VERSION, _T("CzDC v") _T(VERSIONSTRING) _T("[") _T(CZDCVERSIONSTRING) _T("]"));
    SetDlgItemText(IDC_ABOUT_LINK, _T("http://www.czdc.org"));
    
    url.SubclassWindow(GetDlgItem(IDC_ABOUT_LINK));
	url.SetHyperLinkExtendedStyle(HLINK_COMMANDBUTTON|HLINK_UNDERLINEHOVER);
	
    SetDlgItemText(IDC_ABOUT_INFO, _T("Created by Petr \"PPK\" Kozelka\n\nCzDC is DC++ modification licenced under GNU GPL"));
	CEdit ctrl(GetDlgItem(IDC_THANKS));
	ctrl.FmtLines(TRUE);
	ctrl.AppendText(thanks, TRUE);
	ctrl.Detach();
	SetDlgItemText(IDC_TTH, WinUtil::tth.c_str());
	SetDlgItemText(IDC_LATEST, CTSTRING(DOWNLOADING));
		
	char buf[256];
	sprintf_s(buf, 256, "Uptime: %s", WinUtil::formatTime(SETTING(TOTAL_UPTIME)).c_str());
	SetDlgItemText(IDC_TOTAL_UPTIME, Text::toT(buf).c_str());
	sprintf_s(buf, 256, "Upload: %s, Download: %s", Util::formatBytes(SETTING(TOTAL_UPLOAD)).c_str(), Util::formatBytes(SETTING(TOTAL_DOWNLOAD)).c_str());
	SetDlgItemText(IDC_TOTALS, Text::toT(buf).c_str());
	if(SETTING(TOTAL_DOWNLOAD) > 0) {
		sprintf_s(buf, 256, "Ratio (up/down): %.2f", ((double)SETTING(TOTAL_UPLOAD)) / ((double)SETTING(TOTAL_DOWNLOAD)));
		SetDlgItemText(IDC_RATIO, Text::toT(buf).c_str());
	}
	sprintf_s(buf, 256, "Uptime: %s", WinUtil::formatTime(Util::getUptime()).c_str());
	SetDlgItemText(IDC_UPTIME, Text::toT(buf).c_str());
	sprintf_s(buf, 256, "Upload: %s, Download: %s", Util::formatBytes(Socket::getTotalUp()).c_str(), Util::formatBytes(Socket::getTotalDown()).c_str());
	SetDlgItemText(IDC_ACTUALS, Text::toT(buf).c_str());
	if(Socket::getTotalDown() > 0) {
		sprintf_s(buf, 256, "Ratio (up/down): %.2f", ((double)Socket::getTotalUp()) / ((double)Socket::getTotalDown()));
		SetDlgItemText(IDC_ACTUAL_RATIO, Text::toT(buf).c_str());
	}
	CenterWindow(GetParent());
	c.addListener(this);
	c.downloadFile("http://www.czdc.org/version.xml");
	TimerManager::getInstance()->addListener(this);
	return TRUE;
}

LRESULT AboutDlg::onVersionData(UINT /*uMsg*/, WPARAM wParam, LPARAM /*lParam*/, BOOL& /*bHandled*/) {
	tstring* x = (tstring*) wParam;
	SetDlgItemText(IDC_LATEST, x->c_str());
	delete x;
	return 0;
}

LRESULT AboutDlg::OnCloseCmd(WORD /*wNotifyCode*/, WORD wID, HWND /*hWndCtl*/, BOOL& /*bHandled*/) {
	TimerManager::getInstance()->removeListener(this);
	EndDialog(wID);
	return 0;
}

void AboutDlg::on(HttpConnectionListener::Complete, HttpConnection* conn, const string&) throw() {
	if(!downBuf.empty()) {
		SimpleXML xml;
		xml.fromXML(downBuf);
		if(xml.findChild("Version")) {
			string version = xml.getChildAttrib("ActualVersion");
			version += "[" + xml.getChildAttrib("SubVersion") + "]";
			tstring* x = new tstring(Text::toT(version));
			PostMessage(WM_VERSIONDATA, (WPARAM) x);
		}
	}
	conn->removeListener(this);
}

void AboutDlg::on(HttpConnectionListener::Failed, HttpConnection* conn, const string& aLine) throw() {
	tstring* x = new tstring(Text::toT(aLine));
	PostMessage(WM_VERSIONDATA, (WPARAM) x);
	conn->removeListener(this);
}

LRESULT AboutDlg::onSpeaker(UINT /*uMsg*/, WPARAM wParam, LPARAM /*lParam*/, BOOL& /*bHandled*/) {
    static char buf[256];
	if(wParam == STATS_SECOND) {
    	sprintf_s(buf, 256, "Uptime: %s", WinUtil::formatTime(Util::getUptime()).c_str());
    	SetDlgItemText(IDC_UPTIME, Text::toT(buf).c_str());
    	sprintf_s(buf, 256, "Upload: %s, Download: %s", Util::formatBytes(Socket::getTotalUp()).c_str(), Util::formatBytes(Socket::getTotalDown()).c_str());
    	SetDlgItemText(IDC_ACTUALS, Text::toT(buf).c_str());
	} else if(wParam == STATS_MINUTE) {
    	if(Socket::getTotalDown() > 0) {
    		sprintf_s(buf, 256, "Ratio (up/down): %.2f", ((double)Socket::getTotalUp()) / ((double)Socket::getTotalDown()));
    		SetDlgItemText(IDC_ACTUAL_RATIO, Text::toT(buf).c_str());
    	}
    	sprintf_s(buf, 256, "Uptime: %s", WinUtil::formatTime(SETTING(TOTAL_UPTIME)).c_str());
    	SetDlgItemText(IDC_TOTAL_UPTIME, Text::toT(buf).c_str());
    	sprintf_s(buf, 256, "Upload: %s, Download: %s", Util::formatBytes(SETTING(TOTAL_UPLOAD)).c_str(), Util::formatBytes(SETTING(TOTAL_DOWNLOAD)).c_str());
    	SetDlgItemText(IDC_TOTALS, Text::toT(buf).c_str());
    	if(SETTING(TOTAL_DOWNLOAD) > 0) {
    		sprintf_s(buf, 256, "Ratio (up/down): %.2f", ((double)SETTING(TOTAL_UPLOAD)) / ((double)SETTING(TOTAL_DOWNLOAD)));
    		SetDlgItemText(IDC_RATIO, Text::toT(buf).c_str());
    	}
	}
	return 0;
};

LRESULT AboutDlg::onLink(WORD /*wNotifyCode*/, WORD /*wID*/, HWND /*hWndCtl*/, BOOL& /*bHandled*/) {
	WinUtil::openLink(_T("http://www.czdc.org"));
	return 0;
}
