/* 
 * Copyright (C) 2002-2008 Petr Kozelka, PPK at czdc dot org
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA.
 */

#include "stdafx.h"
#include "../client/DCPlusPlus.h"
#include "Resource.h"

#include "CzDCPage.h"
#include "../client/SettingsManager.h"
#include "WinUtil.h"
#include "MainFrm.h"

#ifdef _DEBUG
#define new DEBUG_NEW
#undef THIS_FILE
static char THIS_FILE[] = __FILE__;
#endif

PropPage::TextItem CzDCPage::texts[] = {
	{ IDC_CZDC_FEAT, ResourceManager::SETCZDC_FEAT },
	{ IDC_CZDC_PM_SOUND, ResourceManager::SETCZDC_PRIVATE_SOUND },
	{ IDC_BEEPBROWSE, ResourceManager::BROWSE },
	{ IDC_CZDC_PROGRESS_COLOR, ResourceManager::SETCZDC_PROGRESSBAR_COLORS },
	{ IDC_SETTINGS_DOWNLOAD_BAR_COLOR, ResourceManager::DOWNLOAD },
	{ IDC_PROGRESS_TEXT_COLOR_DOWN, ResourceManager::DOWNLOAD },
	{ IDC_SETTINGS_UPLOAD_BAR_COLOR, ResourceManager::SETCZDC_UPLOAD },
	{ IDC_PROGRESS_TEXT_COLOR_UP, ResourceManager::SETCZDC_UPLOAD },
	{ IDC_CZDC_PROGRESS_TEXT, ResourceManager::SETCZDC_PROGRESSBAR_TEXT },
	{ IDC_CZDC_WINAMP, ResourceManager::SETCZDC_WINAMP },
	{ IDC_SET_SHUTDOWNTIMEOUT, ResourceManager::SETCZDC_SHUTDOWNTIMEOUT },
	{ IDC_CLEARKICKLINE, ResourceManager::CLEARKICKLINE },
	{ IDC_CLEARSEARCHES, ResourceManager::CLEARSEARCHES },
	{ IDC_USERLISTDBLCLICKACTION, ResourceManager::USERLISTDBLCLICKACTION },
	{ IDC_TRANSFERLISTDBLCLICKACTION, ResourceManager::TRANSFERLISTDBLCLICKACTION },	
	{ IDC_CHATDBLCLICKACTION, ResourceManager::CHATDBLCLICKACTION },
	{ 0, ResourceManager::SETTINGS_AUTO_AWAY }
}; 

PropPage::Item CzDCPage::items[] = {
	{ IDC_WINAMP, SettingsManager::WINAMP_FORMAT, PropPage::T_STR },
	{ IDC_SHUTDOWNTIMEOUT, SettingsManager::SHUTDOWN_TIMEOUT, PropPage::T_INT },
	{ 0, 0, PropPage::T_END }
};

LRESULT CzDCPage::onInitDialog(UINT /*uMsg*/, WPARAM /*wParam*/, LPARAM /*lParam*/, BOOL& /*bHandled*/)
{
	PropPage::translate((HWND)(*this), texts);
	PropPage::read((HWND)*this, items);

	crProgressDown = SETTING(DOWNLOAD_BAR_COLOR);
	crProgressUp = SETTING(UPLOAD_BAR_COLOR);
	crProgressTextDown = SETTING(PROGRESS_TEXT_COLOR_DOWN);
	crProgressTextUp = SETTING(PROGRESS_TEXT_COLOR_UP);

	userlistaction.Attach(GetDlgItem(IDC_USERLIST_DBLCLICK));
	transferlistaction.Attach(GetDlgItem(IDC_TRANSFERLIST_DBLCLICK));
	chataction.Attach(GetDlgItem(IDC_CHAT_DBLCLICK));

    userlistaction.AddString(CTSTRING(GET_FILE_LIST));
    userlistaction.AddString(CTSTRING(SEND_PUBLIC_MESSAGE));
    userlistaction.AddString(CTSTRING(SEND_PRIVATE_MESSAGE));
    userlistaction.AddString(CTSTRING(MATCH_QUEUE));
    userlistaction.AddString(CTSTRING(GRANT_EXTRA_SLOT));
	userlistaction.AddString(CTSTRING(ADD_TO_FAVORITES));
	transferlistaction.AddString(CTSTRING(SEND_PRIVATE_MESSAGE));
	transferlistaction.AddString(CTSTRING(GET_FILE_LIST));
	transferlistaction.AddString(CTSTRING(MATCH_QUEUE));
	transferlistaction.AddString(CTSTRING(GRANT_EXTRA_SLOT));
	transferlistaction.AddString(CTSTRING(ADD_TO_FAVORITES));
    chataction.AddString(CTSTRING(SELECT_USER_LIST));
    chataction.AddString(CTSTRING(SEND_PUBLIC_MESSAGE));
    chataction.AddString(CTSTRING(SEND_PRIVATE_MESSAGE));
    chataction.AddString(CTSTRING(GET_FILE_LIST));
    chataction.AddString(CTSTRING(MATCH_QUEUE));
    chataction.AddString(CTSTRING(GRANT_EXTRA_SLOT));
	chataction.AddString(CTSTRING(ADD_TO_FAVORITES));

	userlistaction.SetCurSel(SETTING(USERLIST_DBLCLICK));
	transferlistaction.SetCurSel(SETTING(TRANSFERLIST_DBLCLICK));
	chataction.SetCurSel(SETTING(CHAT_DBLCLICK));

	userlistaction.Detach();
	transferlistaction.Detach();
	chataction.Detach();
	
	// Do specialized reading here
	SetDlgItemText(IDC_BEEPFILE, Text::toT(SETTING(BEEPFILE)).c_str());

	return TRUE;
}

void CzDCPage::write()
{
	PropPage::write((HWND)*this, items);

	// Do specialized writing here
	TCHAR buf[256];
	GetDlgItemText(IDC_BEEPFILE, buf, 256);
	settings->set(SettingsManager::BEEPFILE, Text::fromT(buf));

	SettingsManager::getInstance()->set(SettingsManager::DOWNLOAD_BAR_COLOR, (int)crProgressDown);
	SettingsManager::getInstance()->set(SettingsManager::UPLOAD_BAR_COLOR, (int)crProgressUp);
	SettingsManager::getInstance()->set(SettingsManager::PROGRESS_TEXT_COLOR_DOWN, (int)crProgressTextDown);
	SettingsManager::getInstance()->set(SettingsManager::PROGRESS_TEXT_COLOR_UP, (int)crProgressTextUp);
	userlistaction.Attach(GetDlgItem(IDC_USERLIST_DBLCLICK));
	transferlistaction.Attach(GetDlgItem(IDC_TRANSFERLIST_DBLCLICK));
	chataction.Attach(GetDlgItem(IDC_CHAT_DBLCLICK));
	settings->set(SettingsManager::USERLIST_DBLCLICK, userlistaction.GetCurSel());
	settings->set(SettingsManager::TRANSFERLIST_DBLCLICK, transferlistaction.GetCurSel());
	settings->set(SettingsManager::CHAT_DBLCLICK, chataction.GetCurSel());
	userlistaction.Detach();
	transferlistaction.Detach();
	chataction.Detach(); 
}

LRESULT CzDCPage::onBrowse(WORD /*wNotifyCode*/, WORD /*wID*/, HWND /*hWndCtl*/, BOOL& /*bHandled*/) {
	TCHAR buf[MAX_PATH];

	GetDlgItemText(IDC_BEEPFILE, buf, MAX_PATH);
	tstring x = buf;

	if(WinUtil::browseFile(x, m_hWnd, false) == IDOK) {
		SetDlgItemText(IDC_BEEPFILE, x.c_str());
	}
	return 0;
}

LRESULT CzDCPage::onClickedProgressDown(WORD /* wNotifyCode */, WORD /* wID */, HWND /* hWndCtl */, BOOL& /* bHandled */) {
	CColorDialog d(crProgressDown, 0, *this);
	if (d.DoModal() == IDOK) {
		crProgressDown = d.GetColor();
	}
	return TRUE;
}

LRESULT CzDCPage::onClickedProgressUp(WORD /* wNotifyCode */, WORD /* wID */, HWND /* hWndCtl */, BOOL& /* bHandled */) {
	CColorDialog d(crProgressUp, 0, *this);
	if (d.DoModal() == IDOK) {
		crProgressUp = d.GetColor();
	}
	return TRUE;
}

LRESULT CzDCPage::onClickedProgressTextDown(WORD /* wNotifyCode */, WORD /* wID */, HWND /* hWndCtl */, BOOL& /* bHandled */) {
	CColorDialog d(crProgressTextDown, 0, *this);
	if (d.DoModal() == IDOK) {
		crProgressTextDown = d.GetColor();
	}
	return TRUE;
}

LRESULT CzDCPage::onClickedProgressTextUp(WORD /* wNotifyCode */, WORD /* wID */, HWND /* hWndCtl */, BOOL& /* bHandled */) {
	CColorDialog d(crProgressTextUp, 0, *this);
	if (d.DoModal() == IDOK) {
		crProgressTextUp = d.GetColor();
	}
	return TRUE;
}

LRESULT CzDCPage::onClickedWinampHelp(WORD /* wNotifyCode */, WORD /*wID*/, HWND /* hWndCtl */, BOOL& /* bHandled */) {
	MessageBox(CTSTRING(WINAMP_HELP), CTSTRING(WINAMP_HELP_DESC), MB_OK | MB_ICONINFORMATION);
	return S_OK;
}

LRESULT CzDCPage::onClearKickline(WORD /* wNotifyCode */, WORD /*wID*/, HWND /* hWndCtl */, BOOL& /* bHandled */) {
	SettingsManager::getInstance()->set(SettingsManager::KICK_MSG_RECENT_01, "");
	SettingsManager::getInstance()->set(SettingsManager::KICK_MSG_RECENT_02, "");
	SettingsManager::getInstance()->set(SettingsManager::KICK_MSG_RECENT_03, "");
	SettingsManager::getInstance()->set(SettingsManager::KICK_MSG_RECENT_04, "");
	SettingsManager::getInstance()->set(SettingsManager::KICK_MSG_RECENT_05, "");
	SettingsManager::getInstance()->set(SettingsManager::KICK_MSG_RECENT_06, "");
	SettingsManager::getInstance()->set(SettingsManager::KICK_MSG_RECENT_07, "");
	SettingsManager::getInstance()->set(SettingsManager::KICK_MSG_RECENT_08, "");
	SettingsManager::getInstance()->set(SettingsManager::KICK_MSG_RECENT_09, "");
	SettingsManager::getInstance()->set(SettingsManager::KICK_MSG_RECENT_10, "");
	SettingsManager::getInstance()->set(SettingsManager::KICK_MSG_RECENT_11, "");
	SettingsManager::getInstance()->set(SettingsManager::KICK_MSG_RECENT_12, "");
	SettingsManager::getInstance()->set(SettingsManager::KICK_MSG_RECENT_13, "");
	SettingsManager::getInstance()->set(SettingsManager::KICK_MSG_RECENT_14, "");
	SettingsManager::getInstance()->set(SettingsManager::KICK_MSG_RECENT_15, "");
	SettingsManager::getInstance()->set(SettingsManager::KICK_MSG_RECENT_16, "");
	SettingsManager::getInstance()->set(SettingsManager::KICK_MSG_RECENT_17, "");
	SettingsManager::getInstance()->set(SettingsManager::KICK_MSG_RECENT_18, "");
	SettingsManager::getInstance()->set(SettingsManager::KICK_MSG_RECENT_19, "");
	SettingsManager::getInstance()->set(SettingsManager::KICK_MSG_RECENT_20, "");
	return S_OK;
}

LRESULT CzDCPage::onClearSearches(WORD /* wNotifyCode */, WORD /*wID*/, HWND /* hWndCtl */, BOOL& /* bHandled */) {
	SettingsManager::getInstance()->set(SettingsManager::SEARCH_MSG_RECENT_01, "");
	SettingsManager::getInstance()->set(SettingsManager::SEARCH_MSG_RECENT_02, "");
	SettingsManager::getInstance()->set(SettingsManager::SEARCH_MSG_RECENT_03, "");
	SettingsManager::getInstance()->set(SettingsManager::SEARCH_MSG_RECENT_04, "");
	SettingsManager::getInstance()->set(SettingsManager::SEARCH_MSG_RECENT_05, "");
	SettingsManager::getInstance()->set(SettingsManager::SEARCH_MSG_RECENT_06, "");
	SettingsManager::getInstance()->set(SettingsManager::SEARCH_MSG_RECENT_07, "");
	SettingsManager::getInstance()->set(SettingsManager::SEARCH_MSG_RECENT_08, "");
	SettingsManager::getInstance()->set(SettingsManager::SEARCH_MSG_RECENT_09, "");
	SettingsManager::getInstance()->set(SettingsManager::SEARCH_MSG_RECENT_10, "");
	SettingsManager::getInstance()->set(SettingsManager::SEARCH_MSG_RECENT_11, "");
	SettingsManager::getInstance()->set(SettingsManager::SEARCH_MSG_RECENT_12, "");
	SettingsManager::getInstance()->set(SettingsManager::SEARCH_MSG_RECENT_13, "");
	SettingsManager::getInstance()->set(SettingsManager::SEARCH_MSG_RECENT_14, "");
	SettingsManager::getInstance()->set(SettingsManager::SEARCH_MSG_RECENT_15, "");
	SettingsManager::getInstance()->set(SettingsManager::SEARCH_MSG_RECENT_16, "");
	SettingsManager::getInstance()->set(SettingsManager::SEARCH_MSG_RECENT_17, "");
	SettingsManager::getInstance()->set(SettingsManager::SEARCH_MSG_RECENT_18, "");
	SettingsManager::getInstance()->set(SettingsManager::SEARCH_MSG_RECENT_19, "");
	SettingsManager::getInstance()->set(SettingsManager::SEARCH_MSG_RECENT_20, "");
	MainFrame::clearSearches();
	return S_OK;
}
